@@ -1,6 +1,6 @@
 package Mojo::IOLoop::ReadWriteProcess;
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 
 use Mojo::Base 'Mojo::EventEmitter';
 use Mojo::File 'path';
@@ -17,33 +17,46 @@ use IO::Handle;
 use IO::Pipe;
 use IO::Select;
 use IPC::Open3;
-use POSIX ":sys_wait_h";
+use POSIX qw( :sys_wait_h :signal_h );
 use Symbol 'gensym';
 use Storable;
+use Mojo::IOLoop::Stream;
+use Config;
+
 use constant DEBUG => $ENV{MOJO_PROCESS_DEBUG};
 
+# See https://github.com/torvalds/linux/blob/master/include/uapi/linux/prctl.h
+use constant PR_SET_CHILD_SUBREAPER => 36;
+use constant PR_GET_CHILD_SUBREAPER => 37;
+
+has [
+  qw(kill_sleeptime sleeptime_during_kill),
+  qw(separate_err autoflush set_pipes verbose),
+  qw(internal_pipes collect_status)
+] => 1;
+
+has [qw(blocking_stop serialize detect_subprocess subreaper)] => 0;
+
 has [
   qw(execute code process_id pidfile return_status),
   qw(channel_in channel_out write_stream read_stream error_stream),
   qw(_internal_err _internal_return _status)
 ];
-has blocking_stop         => 0;
-has max_kill_attempts     => 5;
-has kill_sleeptime        => 1;
-has sleeptime_during_kill => 1;
-has args                  => sub { [] };
-has separate_err          => 1;
-has autoflush             => 1;
-has error                 => sub { Mojo::Collection->new };
-has set_pipes             => 1;
-has verbose               => 1;
-has internal_pipes        => 1;
-has serialize             => 0;
-has _deparse              => sub { B::Deparse->new }
+
+has max_kill_attempts => 5;
+
+has args  => sub { [] };
+has error => sub { Mojo::Collection->new };
+
+has subprocess => sub { Mojo::Collection->new };
+has ioloop     => sub { Mojo::IOLoop->singleton };
+
+has _deparse => sub { B::Deparse->new }
   if DEBUG;
 has _deserialize => sub { \&Storable::thaw };
 has _serialize   => sub { \&Storable::freeze };
-has _default_kill_signal => POSIX::SIGTERM;
+has _default_kill_signal     => POSIX::SIGTERM;
+has _default_blocking_signal => POSIX::SIGKILL;
 
 # Override new() just to support sugar syntax
 # so it is possible to do : process->new(sub{ print "Hello World\n" })->start->stop; and so on.
@@ -52,6 +65,20 @@ sub new {
   return shift->SUPER::new(@_);
 }
 
+sub to_ioloop {
+  my $self = shift;
+  confess 'Pipes needs to be set!' unless $self->read_stream;
+  my $stream = Mojo::IOLoop::Stream->new($self->read_stream)->timeout(0);
+  $self->ioloop->stream($stream);
+  my $me = $$;
+  $stream->on(
+    close => sub {
+      return unless $$ == $me;
+      $self->_collect->stop unless defined $self->_status;
+    });
+  return $stream;
+}
+
 sub process { __PACKAGE__->new(@_) }
 sub batch   { Mojo::IOLoop::ReadWriteProcess::Pool->new(@_) }
 sub queue   { Mojo::IOLoop::ReadWriteProcess::Queue->new(@_) }
@@ -65,24 +92,53 @@ sub parallel {
 sub _diag {
   my ($self, @messages) = @_;
   my $caller = (caller(1))[3];
-  print STDERR ">> ${caller}(): @messages\n" if ($self->verbose || DEBUG);
+  print STDERR ">> ${caller}(): @messages\n" if (DEBUG || $self->verbose);
+}
+
+sub _set_child_handler {
+  my $self     = shift;
+  my $sigset   = POSIX::SigSet->new;
+  my $blockset = POSIX::SigSet->new(SIGCHLD);
+
+  sigprocmask(SIG_BLOCK, $blockset, $sigset);
+  $SIG{CHLD} = sub {
+    local ($!, $?);
+    $self->emit('SIG_CHLD');
+    return unless $self->collect_status;
+    while ((my $pid = waitpid(-1, WNOHANG)) > 0) {
+      $self->emit('collect_status' => $pid => $? => $!);
+    }
+  };
+  sigprocmask(SIG_SETMASK, $sigset);
 }
 
 sub _open_collect_status {
-  my $self = shift;
+  my ($self, $pid, $e, $errno) = @_;
+
   return unless $self;
-  $self->_status($?);
+  return push(
+    @{$self->subprocess},
+    $self->new->pid($pid)->_open_collect_status($pid => $e => $errno))
+    && $self->emit(new_subprocess => $self->subprocess->last)
+    if $self->detect_subprocess
+    && defined $pid
+    && $pid != $self->process_id;
+
+  $self->_status($? // $e);
+  $self->_diag("Forked code Process Exit status: " . $self->exit_status)
+    if DEBUG;
+
   $self->_clean_pidfile;
+
+  return $self;
 }
 
 # Use open3 to launch external program.
 sub _open {
   my ($self, @args) = @_;
   $self->_diag('Execute: ' . (join ', ', map { "'$_'" } @args)) if DEBUG;
-  $SIG{CHLD} = sub {
-    local ($!, $?);
-    $self->emit('collect_status') while ((my $pid = waitpid(-1, WNOHANG)) > 0);
-  };
+
+  $self->_set_child_handler;
 
   my ($wtr, $rdr, $err);
   $err = gensym;
@@ -92,7 +148,7 @@ sub _open {
   $self->process_id($pid);
 
   # Defered collect of return status and removal of pidfile
-  $self->once(collect_status => \&_open_collect_status);
+  $self->on(collect_status => \&_open_collect_status);
 
   return $self unless $self->set_pipes();
 
@@ -107,15 +163,34 @@ sub _open {
 
 sub _clean_pidfile { unlink(shift->pidfile) if $_[0]->pidfile }
 
+sub _collect {
+  my ($self, $pid) = @_;
+  $pid //= $self->pid;
+  waitpid $pid, 0;
+  return $self->_open_collect_status($pid) if $self->execute;
+  return $self->_fork_collect_status($pid) if $self->code;
+}
+
 sub _fork_collect_status {
-  my $self = shift;
+  my ($self, $pid, $e, $errno) = @_;
+
   return unless $self;
+  return push(
+    @{$self->subprocess},
+    $self->new->pid($pid)->_fork_collect_status($pid => $e => $errno))
+    && $self->emit(new_subprocess => $self->subprocess->last)
+    if $self->detect_subprocess
+    && defined $pid
+    && $pid != $self->process_id;
+
   my $return_reader;
   my $internal_err_reader;
   my $rt;
   my @result_error;
 
-  $self->_status($?);
+  $self->_status($? // $e);
+  $self->_diag("Forked code Process Exit status: " . $self->exit_status)
+    if DEBUG;
 
   if ($self->_internal_return) {
     $return_reader
@@ -125,7 +200,8 @@ sub _fork_collect_status {
     $self->_new_err('Cannot read from return code pipe') && return
       unless IO::Select->new($return_reader)->can_read(10);
     $rt = $return_reader->getline();
-    $self->_diag("Forked code Process Returns: " . $rt) if DEBUG;
+    $self->_diag("Forked code Process Returns: " . ($rt ? $rt : 'nothing'))
+      if DEBUG;
     $self->return_status($self->serialize ?
         eval { $self->_deserialize->(b64_decode($rt)) }
       : $rt ? $rt
@@ -148,6 +224,7 @@ sub _fork_collect_status {
   }
 
   $self->_clean_pidfile;
+  return $self;
 }
 
 # Handle forking of code
@@ -185,19 +262,17 @@ sub _fork {
   }
 
   # Defered collect of return status
-  $self->once(collect_status => \&_fork_collect_status);
 
-  if (DEBUG) {
-    my $code_str = $self->_deparse->coderef2text($code);
-    $self->_diag("Fork: $code_str");
-  }
+  $self->on(collect_status => \&_fork_collect_status);
+
+  $self->_diag("Fork: " . $self->_deparse->coderef2text($code)) if DEBUG;
 
   my $pid = fork;
   die "Cannot fork: $!" unless defined $pid;
 
   if ($pid == 0) {
     local $SIG{CHLD};
-    local $SIG{TERM} = sub { $self->_exit(1) };
+    local $SIG{TERM} = sub { $self->emit('SIG_TERM')->_exit(1) };
 
     my $return;
     my $internal_err;
@@ -260,7 +335,8 @@ sub _fork {
       $internal_err->write($@) if $@;
       $internal_err->write($!) if !$@ && $!;
     }
-    $rt = @$rt[0] if scalar @$rt == 1 && !$self->serialize;
+    $rt = @$rt[0]
+      if !$self->serialize && ref $rt eq 'ARRAY' && scalar @$rt == 1;
     $rt = b64_encode(eval { $self->_serialize->($rt) })
       if $self->serialize && $return;
     $return->write($rt) if $return;
@@ -268,10 +344,7 @@ sub _fork {
   }
   $self->process_id($pid);
 
-  $SIG{CHLD} = sub {
-    local ($!, $?);
-    $self->emit('collect_status') while ((my $pid = waitpid(-1, WNOHANG)) > 0);
-  };
+  $self->_set_child_handler;
 
   return $self unless $self->set_pipes();
 
@@ -295,8 +368,7 @@ sub _new_err {
 
   # XXX: Need to switch, we should emit one error at the time, and _shutdown
   # should emit just the ones wasn't emitted
-  $self->emit(process_error => [$err]);
-  return $self;
+  return $self->emit(process_error => [$err]);
 }
 
 sub _exit {
@@ -388,6 +460,8 @@ sub start {
       : $self->args
     : ();
 
+  $self->enable_subreaper if $self->subreaper;
+
   if ($self->code) {
     $self->_fork($self->code, @args);
   }
@@ -401,8 +475,18 @@ sub start {
   return $self;
 }
 
-sub signal {
-  my $self = shift;
+sub disable_subreaper {
+  $_[0]->subreaper(1) unless $_[0]->_prctl(PR_SET_CHILD_SUBREAPER, 0) == 0;
+  $_[0];
+}
+
+sub enable_subreaper {
+  $_[0]->subreaper(0) unless $_[0]->_prctl(PR_SET_CHILD_SUBREAPER, 1) == 0;
+  $_[0];
+}
+
+sub send_signal {
+  my $self   = shift;
   my $signal = shift // $self->_default_kill_signal;
   return unless $self->is_running;
   $self->_diag("Sending signal '$signal' to " . $self->process_id) if DEBUG;
@@ -412,7 +496,7 @@ sub signal {
 
 sub stop {
   my $self = shift;
-  return $self->_shutdown unless $self->is_running;
+  return $self->_shutdown(1) unless $self->is_running;
 
   my $ret;
   my $attempt = 1;
@@ -426,7 +510,7 @@ sub stop {
         . $self->pid)
       if DEBUG;
     sleep $self->sleeptime_during_kill if $self->sleeptime_during_kill;
-    $self->signal();
+    $self->send_signal();
     $ret = waitpid($self->process_id, WNOHANG);
     $self->_status($?);
     $attempt++;
@@ -436,9 +520,9 @@ sub stop {
 
   if ($self->blocking_stop && $self->is_running) {
     $self->_diag(
-      "Could not kill process id: " . $self->process_id . " going for SIGKILL")
+      "Could not kill process id: " . $self->process_id . ", blocking attempt")
       if DEBUG;
-    $self->signal(POSIX::SIGKILL);
+    $self->emit('process_stuck')->send_signal($self->_default_blocking_signal);
     waitpid($self->process_id, 0);
     $self->_status($?);
   }
@@ -452,21 +536,85 @@ sub stop {
 
 sub _shutdown {
   my $self = shift;
+  waitpid $self->pid, 0 if pop;
   $self->emit('collect_status') if !defined $self->_status;
   $self->_clean_pidfile;
   $self->emit('process_error', $self->error)
     if $self->error && $self->error->size > 0;
-  $self->emit('stop');
-  return $self;
+  $self->unsubscribe('collect_status');
+  return $self->emit('stop');
 }
 
-sub DESTROY { +shift()->_shutdown; }
+
+sub _get_prctl_syscall {
+
+  # Courtesy of Sys::Prctl
+  confess "Only Linux is supported" unless $^O eq 'linux';
+
+  my $machine = (POSIX::uname())[4];
+  die "Could not get machine type" unless $machine;
+
+  # if we're running on an x86_64 kernel, but a 32-bit process,
+  # we need to use the i386 syscall numbers.
+  $machine = "i386" if ($machine eq "x86_64" && $Config{ptrsize} == 4);
+
+  my $prctl_call
+    = $machine
+    =~ /^i[3456]86|^blackfin|cris|frv|h8300|m32r|m68k|microblaze|mn10300|sh|s390|parisc$/
+    ? 172
+    : $machine eq "x86_64"  ? 157
+    : $machine eq "sparc64" ? 147
+    : $machine eq "ppc"     ? 171
+    : $machine eq "ia64"    ? 1170
+    : $machine eq "alpha"   ? 348
+    : $machine eq "arm"     ? 0x900000 + 172
+    : $machine eq "avr32"   ? 148
+    : $machine eq "mips"    ? 4000 + 192
+    : $machine eq "mips64"  ? 5000 + 153
+    : $machine eq "xtensa"  ? 130
+    :                         undef;
+
+  unless (defined $prctl_call) {
+    delete @INC{
+      qw<syscall.ph asm/unistd.ph bits/syscall.ph _h2ph_pre.ph
+        sys/syscall.ph>
+    };
+    my $rv = eval { require 'syscall.ph'; 1 }     ## no critic
+      or eval { require 'sys/syscall.ph'; 1 };    ## no critic
+
+    $prctl_call = eval { &SYS_prctl; };
+  }
+  _diag(undef, "prctl() $prctl_call") if DEBUG;
+
+  return $prctl_call;
+}
+
+sub _prctl {
+  my ($self, $option, $arg2, $arg3, $arg4, $arg5) = @_;
+  confess 'prctl not supported in this platform!'
+    unless defined _get_prctl_syscall;
+  local $!;
+  my $ret = syscall(
+    _get_prctl_syscall(), $option,
+    ($arg2 or 0),
+    ($arg3 or 0),
+    ($arg4 or 0),
+    ($arg5 or 0));
+
+  $self->_diag($self->_prctl_call, "$option is unavailable on this platform.")
+    if $!{EINVAL};
+  $self->_diag("Error! $!") if $!;
+  return $ret;
+}
 
 # General alias
-*pid  = \&process_id;
-*died = \&_errored;
-*diag = \&_diag;
-*pool = \&batch;
+*pid    = \&process_id;
+*died   = \&_errored;
+*failed = \&_errored;
+*diag   = \&_diag;
+*pool   = \&batch;
+*signal = \&send_signal;
+*prctl  = \&_prctl;
 
 # Aliases - write
 *write         = \&write_stdin;
@@ -555,26 +703,6 @@ Mojo::IOLoop::ReadWriteProcess is yet another process manager.
 L<Mojo::IOLoop::ReadWriteProcess> inherits all events from L<Mojo::EventEmitter> and can emit
 the following new ones.
 
-=head2 collect_status
-
- $process->on(collect_status => sub {
-   my ($self) = @_;
-   ...
- });
-
-Emitted when on child process waitpid.
-It is used internally to get the child process status.
-
-=head2 process_error
-
- $process->on(process_error => sub {
-   my ($e) = @_;
-   my @errors = @{$e};
- });
-
-Emitted when the process produce errors.
-
-
 =head2 start
 
  $process->on(start => sub {
@@ -593,6 +721,65 @@ Emitted when the process starts.
 
 Emitted when the process stops.
 
+=head2 process_error
+
+ $process->on(process_error => sub {
+   my ($e) = @_;
+   my @errors = @{$e};
+ });
+
+Emitted when the process produce errors.
+
+=head2 process_stuck
+
+ $process->on(process_stuck => sub {
+   my ($self) = @_;
+   ...
+ });
+
+Emitted when C<blocking_stop> is set and all attempts for killing the process
+in C<max_kill_attempts> have been exhausted.
+The event is emitted before attempting to kill it with SIGKILL and becoming blocking.
+
+=head2 new_subprocess
+
+ $process->on(new_subprocess => sub {
+   my ($self, $new_process) = @_;
+   my $status = $new_process->exit_status;
+   ...
+ });
+
+Emitted when C<detect_subprocess> is set and will be fired with the collected process.
+
+=head2 SIG_CHLD
+
+ $process->on(SIG_CHLD => sub {
+   my ($self) = @_;
+   ...
+ });
+
+Emitted when we receive SIG_CHLD.
+
+=head2 SIG_TERM
+
+ $process->on(SIG_TERM => sub {
+   my ($self) = @_;
+   ...
+ });
+
+Emitted when the child forked process receives SIG_TERM, before exiting.
+
+=head2 collect_status
+
+ $process->on(collect_status => sub {
+   my ($self) = @_;
+   ...
+ });
+
+Emitted when on child process waitpid.
+It is used internally to get the child process status.
+Note: events attached to it are wiped when process has been stopped.
+
 =head1 ATTRIBUTES
 
 L<Mojo::IOLoop::ReadWriteProcess> inherits all attributes from L<Mojo::EventEmitter> and implements
@@ -647,6 +834,67 @@ Array or arrayref of options to pass by to the external binary or the code block
 
 Set it to 1 if you want to do blocking stop of the process.
 
+=head2 detect_subprocess
+
+    use Mojo::IOLoop::ReadWriteProcess qw(process);
+
+    # fork as usual in the code, or either create processes and never wait for them.
+    process(sub { ... expensive operation ... } )->start();
+
+    my $process = process(sub {});
+    $process->detect_subprocess(1);
+    $process->on( new_subprocess => sub { print "Not known process: ".(pop->pid)." reaped" } );
+    $process->start();
+
+    # Move other bits..
+
+    $process->wait_stop(); # Just wait on the master one (must be last one called in sequence)
+
+    my @sub_p = $process->subprocess()->to_array;
+
+Set it to 1 if you want the last started process to be master and be able to
+detect ended processes.
+
+=head2 subreaper
+
+    use Mojo::IOLoop::ReadWriteProcess;
+    my $process = Mojo::IOLoop::ReadWriteProcess->new(code => sub { print "Hello ".shift() }, args => "User" );
+    $process->detect_subprocess(1)->subreaper(1)->start();
+    $process->on( stop => sub { $_->disable_subreaper } );
+    $process->stop();
+
+    # The process will print "Hello User"
+
+Mark the current process (not the child) as subreaper on start.
+It's on invoker behalf to disable subreaper when process stops, as it marks the current process and not the
+child.
+
+=head2 subprocess
+
+    use Mojo::IOLoop::ReadWriteProcess qw(process);
+
+    # fork as usual in the code, or either create processes and never wait for them.
+    process(sub { ... expensive operation ... } )->start();
+
+    my $process = process(sub { print "Hello"; sleep 1 });
+    $process->detect_subprocess(1);
+    $process->start();
+    $process->on( new_subprocess => sub { print "Not known process: ".(pop->pid)." reaped" } );
+    $process->wait_stop(); # Just wait on the master one (must be last one called in sequence)
+
+    my $subprocesses = $process->subprocess();
+    my $first = $subprocesses->first;
+
+Returns a L<Mojo::Collection> of L<Mojo::IOLoop::ReadWriteProcess> that contains the pid and
+the exit status of other processes that are sending a SIG_CHLD.
+
+=head2 ioloop
+
+  my $loop    = $process->ioloop;
+  $subprocess = $process->ioloop(Mojo::IOLoop->new);
+
+Event loop object to control, defaults to the global L<Mojo::IOLoop> singleton.
+
 =head2 max_kill_attempts
 
     use Mojo::IOLoop::ReadWriteProcess;
@@ -660,6 +908,13 @@ Defaults to C<5>, is the number of attempts before bailing out.
 It can be used with blocking_stop, so if the number of attempts are exhausted,
 a SIGKILL and waitpid will be tried at the end.
 
+=head2 collect_status
+
+Defaults to C<1>, If enabled it will automatically collect the status of the children process.
+Disable it in case you want to manage your process child directly, and do not want to rely on
+automatic collect status. If you won't overwrite your C<SIGCHLD> handler,
+the C<SIG_CHLD> event will be still emitted.
+
 =head2 serialize
 
 Defaults to C<0>, If enabled data returned from forked process will be serialized with Storable.
@@ -756,6 +1011,58 @@ call C<_status()>.
 
 Inspect the codeblock return.
 
+=head2 enable_subreaper()
+
+    use Mojo::IOLoop::ReadWriteProcess qw(process);
+    my $p = process()->enable_subreaper;
+
+Mark the current process (not the child) as subreaper.
+This is used typically if you want to mark further childs as subreapers inside other forks.
+
+    my $master_p = process(
+      sub {
+        my $p = shift;
+        $p->enable_subreaper;
+
+        process(sub { sleep 4; exit 1 })->start();
+        process(
+          sub {
+            sleep 4;
+            process(sub { sleep 1; })->start();
+          })->start();
+        process(sub { sleep 4; exit 0 })->start();
+        process(sub { sleep 4; die })->start();
+        my $manager
+          = process(sub { sleep 2 })->detect_subprocess(1)->subreaper(1)->start();
+        sleep 1 for (0 .. 10);
+        $manager->stop;
+        return $manager->subprocess->size;
+      });
+
+    $master_p->detect_subprocess(1);
+    $master_p->subreaper(1);
+    $master_p->on(collect_status => sub { $status++ });
+
+    $master_p->on(stop => sub { shift()->disable_subreaper });
+    $master_p->start();
+
+    ....
+
+=head2 disable_subreaper()
+
+    use Mojo::IOLoop::ReadWriteProcess qw(process);
+    my $p = process()->disable_subreaper;
+
+Unset the current process (not the child) as subreaper.
+
+=head2 prctl()
+
+    use Mojo::IOLoop::ReadWriteProcess qw(process);
+    my $p = process();
+    $p->prctl($option, $arg2, $arg3, $arg4, $arg5);
+
+Internal function to execute and wrap the prctl syscall, accepts the same arguments as prctl.
+
 =head2 diag()
 
     use Mojo::IOLoop::ReadWriteProcess qw(process);
@@ -766,6 +1073,24 @@ Inspect the codeblock return.
 Internal function to print information to STDERR if verbose attribute is set or either DEBUG mode enabled.
 You can use it if you wish to display information on the process status.
 
+=head2 to_ioloop()
+
+    use Mojo::IOLoop::ReadWriteProcess qw(process);
+
+    my $p = process(sub {  print "Hello from first process\n"; sleep 1 });
+
+    $p->start(); # Start and sets the handlers
+    my $stream = $p->to_ioloop; # Get the stream and demand to IOLoop
+    my $output;
+
+    # Hook on Mojo::IOLoop::Stream events
+    $stream->on(read => sub { $output .= pop;  $p->is_running ...  });
+
+    Mojo::IOLoop->singleton->start() unless Mojo::IOLoop->singleton->is_running;
+
+Returns a L<Mojo::IOLoop::Stream> object and demand the wait operation to L<Mojo::IOLoop>.
+It needs C<set_pipes> enabled. Default IOLoop can be overridden in C<ioloop()>.
+
 =head2 wait()
 
     use Mojo::IOLoop::ReadWriteProcess qw(process);
@@ -903,13 +1228,13 @@ Gets all the channel output of the process.
 
 Gets all the STDERR output of the process.
 
-=head2 signal()
+=head2 send_signal()
 
     use Mojo::IOLoop::ReadWriteProcess qw(process);
     use POSIX;
     my $p = process( execute => "/path/to/bin" )->start;
 
-    $p->signal(POSIX::SIGKILL);
+    $p->send_signal(POSIX::SIGKILL);
 
 Send a signal to the process
 
